@@ -9,17 +9,16 @@ from pants.backend.core.tasks.task import Task
 from pants.backend.android.distribution.android_distribution import AndroidDistribution
 
 class AndroidTask(Task):
-
   # The flag for release/debug will eventually go here (as we build out the ops in other tasks)
   @classmethod
-  def setup_parser(cls, option_group, args, mkflag):
-    option_group.add_option(mkflag('sdk-path'), dest='sdk_path', type='string',
-                            help='Specify a file location and force the Android SDK there to '
-                                 'be used during Android tasks.')
+  def register_options(cls, register):
+    super(AndroidTask, cls).register_options(register)
+    register('--sdk-path',
+             help='Use the Android SDK at this path.')
 
   def __init__(self, *args, **kwargs):
     super(AndroidTask, self).__init__(*args, **kwargs)
-    self.forced_sdk = self.context.options.sdk_path or None
+    self.forced_sdk = self.get_options().sdk_path or None
     self._android_sdk = None
 
   @property
