@@ -7,7 +7,6 @@ from __future__ import (nested_scopes, generators, division, absolute_import, wi
 
 import logging
 
-from pants.base.payload import EmptyPayload
 from pants.base.target import Target
 
 
@@ -21,12 +20,11 @@ class Dependencies(Target):
   NB: This class is commonly referred to by the alias 'target' in BUILD files.
   """
 
-  def __init__(self, *args, **kwargs):
-    super(Dependencies, self).__init__(payload=EmptyPayload(), *args, **kwargs)
-
 
 class DeprecatedDependencies(Dependencies):
   """A subclass for Dependencies that warns that the 'dependencies' alias is deprecated."""
   def __init__(self, *args, **kwargs):
-    logger.warn("""For %s : The alias 'dependencies(..)' has been deprecated in favor of 'target(..)'""" % kwargs['address'].spec)
+    logger.warn("For {0} : The alias 'dependencies(..)' has been deprecated in favor of "
+                "'target(..)'"
+                .format(kwargs['address'].spec))
     super(DeprecatedDependencies, self).__init__(*args, **kwargs)
