@@ -14,12 +14,14 @@ import os
 import pkgutil
 import shutil
 import sys
+import traceback
 
 from twitter.common.collections import OrderedDict, OrderedSet
 from twitter.common.config import Properties
 from twitter.common.log.options import LogOptions
 
-from pants.backend.core.tasks.scm_publish import Namedver, ScmPublish, Semver, Version
+from pants.scm.scm import Scm
+from pants.backend.core.tasks.scm_publish import Namedver, ScmPublish, Semver
 from pants.backend.jvm.ivy_utils import IvyUtils
 from pants.backend.jvm.targets.jarable import Jarable
 from pants.backend.jvm.targets.scala_library import ScalaLibrary
@@ -35,7 +37,8 @@ from pants.base.generator import Generator, TemplateData
 from pants.base.target import Target
 from pants.ivy.bootstrapper import Bootstrapper
 from pants.ivy.ivy import Ivy
-from pants.util.dirutil import safe_mkdir, safe_open, safe_rmtree, touch
+from pants.util.dirutil import safe_mkdir, safe_open, safe_rmtree
+from pants.util.strutil import ensure_text
 
 
 class PushDb(object):
@@ -80,8 +83,13 @@ class PushDb(object):
       """Returns a clone of this entry with the given sha and fingerprint."""
       return PushDb.Entry(self.sem_ver, self.named_ver, self.named_is_latest, sha, fingerprint)
 
-  def __init__(self, props):
-    self._props = props
+    def __repr__(self):
+      return '<%s, %s, %s, %s, %s, %s>' % (
+        self.__class__.__name__, self.sem_ver, self.named_ver, self.named_is_latest,
+        self.sha, self.fingerprint)
+
+  def __init__(self, props=None):
+    self._props = props or OrderedDict()
 
   def get_entry(self, target):
     """Given an internal target, return a PushDb.Entry, which might contain defaults."""
@@ -90,9 +98,9 @@ class PushDb(object):
     major = int(db_get('revision.major', '0'))
     minor = int(db_get('revision.minor', '0'))
     patch = int(db_get('revision.patch', '0'))
-    snapshot = db_get('revision.snapshot', 'false').lower() == 'true'
+    snapshot = str(db_get('revision.snapshot', 'false')).lower() == 'true'
     named_version = db_get('revision.named_version', None)
-    named_is_latest = db_get('revision.named_is_latest', False)
+    named_is_latest = str(db_get('revision.named_is_latest', 'false')).lower() == 'true'
     sha = db_get('revision.sha', None)
     fingerprint = db_get('revision.fingerprint', None)
     sem_ver = Semver(major, minor, patch, snapshot=snapshot)
@@ -108,7 +116,7 @@ class PushDb(object):
     db_set('revision.snapshot', str(pe.sem_ver.snapshot).lower())
     if pe.named_ver:
       db_set('revision.named_version', pe.named_ver.version())
-    db_set('revision.named_is_latest', pe.named_is_latest)
+    db_set('revision.named_is_latest', str(pe.named_is_latest).lower())
     db_set('revision.sha', pe.sha)
     db_set('revision.fingerprint', pe.fingerprint)
 
@@ -149,7 +157,7 @@ class DependencyWriter(object):
     self.template_package_name = template_package_name or __name__
     self.template_relpath = template_relpath
 
-  def write(self, target, path, confs=None):
+  def write(self, target, path, confs=None, extra_confs=None):
     # TODO(John Sirois): a dict is used here to de-dup codegen targets which have both the original
     # codegen target - say java_thrift_library - and the synthetic generated target (java_library)
     # Consider reworking codegen tasks to add removal of the original codegen targets when rewriting
@@ -171,12 +179,12 @@ class DependencyWriter(object):
                                   configurations=list(configurations))
     target_jar = target_jar.extend(dependencies=dependencies.values())
 
-    template_kwargs = self.templateargs(target_jar, confs)
+    template_kwargs = self.templateargs(target_jar, confs, extra_confs)
     with safe_open(path, 'w') as output:
       template = pkgutil.get_data(self.template_package_name, self.template_relpath)
       Generator(template, **template_kwargs).write(output)
 
-  def templateargs(self, target_jar, confs=None):
+  def templateargs(self, target_jar, confs=None, extra_confs=None):
     """
       Subclasses must return a dict for use by their template given the target jar template data
       and optional specific ivy configurations.
@@ -208,7 +216,7 @@ class PomWriter(DependencyWriter):
         get_db,
         os.path.join('templates', 'jar_publish', 'pom.mustache'))
 
-  def templateargs(self, target_jar, confs=None):
+  def templateargs(self, target_jar, confs=None, extra_confs=None):
     return dict(artifact=target_jar)
 
   def jardep(self, jar):
@@ -234,9 +242,10 @@ class IvyWriter(DependencyWriter):
         IvyUtils.IVY_TEMPLATE_PATH,
         template_package_name=IvyUtils.IVY_TEMPLATE_PACKAGE_NAME)
 
-  def templateargs(self, target_jar, confs=None):
+  def templateargs(self, target_jar, confs=None, extra_confs=None):
     return dict(lib=target_jar.extend(
-        publications=set(confs) if confs else set(),
+        publications=set(confs or []),
+        extra_publications=extra_confs if extra_confs else {},
         overrides=None))
 
   def _jardep(self, jar, transitive=True, configurations='default'):
@@ -273,7 +282,18 @@ def pushdb_coordinate(jar, entry):
 
 
 def target_internal_dependencies(target):
-  return filter(lambda tgt: isinstance(tgt, Jarable), target.dependencies)
+  """Returns internal Jarable dependencies that were "directly" declared.
+
+  Directly declared deps are those that are explicitly listed in the definition of a
+  target, rather than being depended on transitively. But in order to walk through
+  aggregator targets such as `target`, `dependencies`, or `jar_library`, this recursively
+  descends the dep graph and stops at Jarable instances."""
+  for dep in target.dependencies:
+    if isinstance(dep, Jarable):
+      yield dep
+    else:
+      for childdep in target_internal_dependencies(dep):
+        yield childdep
 
 
 class JarPublish(JarTask, ScmPublish):
@@ -287,12 +307,14 @@ class JarPublish(JarTask, ScmPublish):
   primarily provide linkage between publishable targets and the
   Ivy ``resolvers`` used to publish them.
 
-  The following target types are publishable: :ref:`bdict_java_library`,
-  :ref:`bdict_scala_library`, :ref:`bdict_java_thrift_library`,
-  :ref:`bdict_annotation_processor`.
+  The following target types are publishable:
+  `java_library <build_dictionary.html#java_library>`_,
+  `scala_library <build_dictionary.html#scala_library>`_,
+  `java_thrift_library <build_dictionary.html#java_thrift_library>`_,
+  `annotation_processor <build_dictionary.html#annotation_processor>`_.
   Targets to publish and their dependencies must be publishable target
   types and specify the ``provides`` argument. One exception is
-  :ref:`bdict_jar`\s - pants will generate a pom file that
+  `jar <build_dictionary.html#jar>`_\s - pants will generate a pom file that
   depends on the already-published jar.
 
   Example usage: ::
@@ -335,10 +357,11 @@ class JarPublish(JarTask, ScmPublish):
   """
 
   _CONFIG_SECTION = 'jar-publish'
+  _SCM_PUSH_ATTEMPTS = 5
 
   @classmethod
-  def setup_parser(cls, option_group, args, mkflag):
-    super(JarPublish, cls).setup_parser(option_group, args, mkflag)
+  def register_options(cls, register):
+    super(JarPublish, cls).register_options(register)
 
     # TODO(John Sirois): Support a preview mode that outputs a file with entries like:
     # artifact id:
@@ -349,89 +372,55 @@ class JarPublish(JarTask, ScmPublish):
     # Allow re-running this goal with the file as input to support forcing an arbitrary set of
     # revisions and supply of hand edited changelogs.
 
-    option_group.add_option(mkflag("dryrun"), mkflag("dryrun", negate=True),
-                            dest="jar_publish_dryrun", default=True,
-                            action="callback", callback=mkflag.set_bool,
-                            help="[%default] Runs through a push without actually pushing "
-                                 "artifacts, editing publish dbs or otherwise writing data")
-
-    option_group.add_option(mkflag("commit", negate=True),
-                            dest="jar_publish_commit", default=True,
-                            action="callback", callback=mkflag.set_bool,
-                            help="Turns off commits of the push db for local testing.")
-
-    local_flag = mkflag("local")
-    option_group.add_option(local_flag, dest="jar_publish_local",
-                            help="Publishes jars to a maven repository on the local filesystem at "
-                                 "the specified path.")
-
-    local_snapshot_flag = mkflag("local-snapshot")
-    option_group.add_option(local_snapshot_flag, mkflag("local-snapshot", negate=True),
-                            dest="jar_publish_local_snapshot", default=True,
-                            action="callback", callback=mkflag.set_bool,
-                            help="[%default] Iff {0} is specified, publishes jars with '-SNAPSHOT' "
-                                 "revision suffixes.".format(local_flag))
-
-    named_snapshot_flag = mkflag("named-snapshot")
-    option_group.add_option(named_snapshot_flag,
-                            dest="jar_publish_named_snapshot",
-                            default=None,
-                            help="Publishes all artifacts with the given snapshot name replacing "
-                                 "their version. This is not Semantic Versioning compatible, but "
-                                 "is easier to consume in cases where many artifacts must align.")
-
-    option_group.add_option(mkflag("transitive"), mkflag("transitive", negate=True),
-                            dest="jar_publish_transitive", default=True,
-                            action="callback", callback=mkflag.set_bool,
-                            help="[%default] Publishes the specified targets and all their "
-                                 "internal dependencies transitively.")
-
-    option_group.add_option(mkflag("force"), mkflag("force", negate=True),
-                            dest="jar_publish_force", default=False,
-                            action="callback", callback=mkflag.set_bool,
-                            help="[%default] Forces pushing jars even if there have been no "
-                                 "changes since the last push.")
-
-    override_flag = mkflag('override')
-    option_group.add_option(override_flag, action='append', dest='jar_publish_override',
-                            help='''Specifies a published jar revision override in the form:
-                            ([org]#[name]|[target spec])=[new revision]
-
-                            For example, to specify 2 overrides:
-                            {flag}=com.twitter.common#quantity=0.1.2 \\
-                            {flag}=src/java/com/twitter/common/base=1.0.0 \\
-                            '''.format(flag=override_flag))
-
-    flag = mkflag("restart-at")
-    option_group.add_option(flag, dest="jar_publish_restart_at",
-                            help='''Restart a fail push at the given jar.  Jars can be identified by
-                            maven coordinate [org]#[name] or target.
-
-                            For example:
-                            %(flag)s=com.twitter.common#quantity
-
-                            Or:
-                            %(flag)s=src/java/com/twitter/common/base
-                            ''' % dict(flag=flag))
+    register('--dryrun', default=True, action='store_true',
+             help='Run through a push without actually pushing artifacts, editing publish dbs or '
+                  'otherwise writing data')
+    register('--commit', default=True, action='store_true',
+             help='Commit the push db. Turn off for local testing.')
+    register('--local', metavar='<PATH>',
+             help='Publish jars to a maven repository on the local filesystem at this path.')
+    register('--scm-push-attempts', type=int, default=cls._SCM_PUSH_ATTEMPTS,
+             help='Try pushing the pushdb to the SCM this many times before aborting.')
+    register('--local-snapshot', default=True, action='store_true',
+             help='If --local is specified, publishes jars with -SNAPSHOT revision suffixes.')
+    register('--named-snapshot', default=None,
+             help='Publish all artifacts with the given snapshot name, replacing their version. '
+                  'This is not Semantic Versioning compatible, but is easier to consume in cases '
+                  'where many artifacts must align.')
+    register('--transitive', default=True, action='store_true',
+             help='Publish the specified targets and all their internal dependencies transitively.')
+    register('--force', default=False, action='store_true',
+             help='Force pushing jars even if there have been no changes since the last push.')
+    register('--override', action='append',
+             help='Specifies a published jar revision override in the form: '
+                  '([org]#[name]|[target spec])=[new revision] '
+                  'For example, to specify 2 overrides: '
+                  '--override=com.foo.bar#baz=0.1.2  --override=src/java/com/foo/bar/qux=1.0.0')
+    register('--restart-at',
+             help='Restart a fail push at the given jar.  Jars can be identified by '
+                  'maven coordinate [org]#[name] or target. '
+                  'For example: --restart-at=com.twitter.common#quantity '
+                  'Or: --restart-at=src/java/com/twitter/common/base')
 
   def __init__(self, *args, **kwargs):
     super(JarPublish, self).__init__(*args, **kwargs)
     ScmPublish.__init__(self, get_scm(),
-                        self.context.config.getlist(self._CONFIG_SECTION, 'restrict_push_branches'))
+                        self.context.config.getlist(self._CONFIG_SECTION,
+                                                    'restrict_push_branches'))
     self.cachedir = os.path.join(self.workdir, 'cache')
 
     self._jvmargs = self.context.config.getlist(self._CONFIG_SECTION, 'ivy_jvmargs', default=[])
 
-    if self.context.options.jar_publish_local:
+    if self.get_options().local:
       local_repo = dict(
         resolver='publish_local',
-        path=os.path.abspath(os.path.expanduser(self.context.options.jar_publish_local)),
+        path=os.path.abspath(os.path.expanduser(self.get_options().local)),
         confs=['default'],
         auth=None
       )
       self.repos = defaultdict(lambda: local_repo)
       self.commit = False
-      self.local_snapshot = self.context.options.jar_publish_local_snapshot
+      self.local_snapshot = self.get_options().local_snapshot
     else:
       self.repos = self.context.config.getdict(self._CONFIG_SECTION, 'repos')
       if not self.repos:
@@ -447,18 +436,18 @@ class JarPublish(JarTask, ScmPublish):
           self.context.log.debug('Found auth for repo=%s user=%s' % (repo, user))
           self.repos[repo]['username'] = user
           self.repos[repo]['password'] = password
-      self.commit = self.context.options.jar_publish_commit
+      self.commit = self.get_options().commit
       self.local_snapshot = False
 
-    self.named_snapshot = self.context.options.jar_publish_named_snapshot
+    self.named_snapshot = self.get_options().named_snapshot
     if self.named_snapshot:
       self.named_snapshot = Namedver.parse(self.named_snapshot)
 
     self.ivycp = self.context.config.getlist('ivy', 'classpath')
 
-    self.dryrun = self.context.options.jar_publish_dryrun
-    self.transitive = self.context.options.jar_publish_transitive
-    self.force = self.context.options.jar_publish_force
+    self.dryrun = self.get_options().dryrun
+    self.transitive = self.get_options().transitive
+    self.force = self.get_options().force
 
     def parse_jarcoordinate(coordinate):
       components = coordinate.split('#', 1)
@@ -484,10 +473,9 @@ class JarPublish(JarTask, ScmPublish):
           .format(message=e, coordinate=coordinate))
 
     self.overrides = {}
-    if self.context.options.jar_publish_override:
+    if self.get_options().override:
       if self.named_snapshot:
-        raise TaskError('Flags {0} and {1} are mutually exclusive!'.format(named_snapshot_flag,
-                                                                           override_flag))
+        raise TaskError('Options --named-snapshot and --override are mutually exclusive!')
 
       def parse_override(override):
         try:
@@ -501,11 +489,11 @@ class JarPublish(JarTask, ScmPublish):
         except ValueError:
           raise TaskError('Invalid override: %s' % override)
 
-      self.overrides.update(parse_override(o) for o in self.context.options.jar_publish_override)
+      self.overrides.update(parse_override(o) for o in self.get_options().override)
 
     self.restart_at = None
-    if self.context.options.jar_publish_restart_at:
-      self.restart_at = parse_jarcoordinate(self.context.options.jar_publish_restart_at)
+    if self.get_options().restart_at:
+      self.restart_at = parse_jarcoordinate(self.get_options().restart_at)
 
   @property
   def config_section(self):
@@ -516,6 +504,90 @@ class JarPublish(JarTask, ScmPublish):
     round_manager.require('javadoc')
     round_manager.require('scaladoc')
 
+  def confirm_push(self, coord, version):
+    """Ask the user if a push should be done for a particular version of a
+       particular coordinate.   Return True if the push should be done"""
+    try:
+      isatty = os.isatty(sys.stdin.fileno())
+    except ValueError:
+      # In tests, sys.stdin might not have a fileno
+      isatty = False
+    if not isatty:
+      return True
+    push = raw_input('Publish %s with revision %s ? [y|N] ' % (
+      coord, version
+    ))
+    print('\n')
+    return push.strip().lower() == 'y'
+
+  def _copy_artifact(self, tgt, jar, version, typename, suffix='', extension='jar',
+                     artifact_ext='', override_name=None):
+    """Copy the products for a target into the artifact path for the jar/version"""
+    genmap = self.context.products.get(typename)
+    product_mapping = genmap.get(tgt)
+    if product_mapping is None:
+      raise ValueError("No product mapping in %s for %s. "
+                       "You may need to run some other task first" % (typename, tgt))
+    for basedir, jars in product_mapping.items():
+      for artifact in jars:
+        path = self.artifact_path(jar, version, name=override_name, suffix=suffix,
+                                  extension=extension, artifact_ext=artifact_ext)
+        safe_mkdir(os.path.dirname(path))
+        shutil.copy(os.path.join(basedir, artifact), path)
+
+  def _ivy_jvm_args(self, repo):
+    """Get the JVM arguments for ivy authentication, if needed"""
+    # Get authentication for the publish repo if needed
+    if not repo.get('auth'):
+      return self._jvmargs
+
+    jvm_args = self._jvmargs
+    user = repo.get('username')
+    password = repo.get('password')
+    if user and password:
+      jvm_args.append('-Dlogin=%s' % user)
+      jvm_args.append('-Dpassword=%s' % password)
+    else:
+      raise TaskError('Unable to publish to %s. %s' %
+                      (repo.get('resolver'), repo.get('help', '')))
+    return jvm_args
+
+  def publish(self, ivyxml_path, jar, entry, repo, published):
+    """Run ivy to publish a jar.  ivyxml_path is the path to the ivy file; published
+    is a list of jars published so far (including this one). entry is a pushdb entry."""
+    jvm_args = self._ivy_jvm_args(repo)
+    resolver = repo['resolver']
+    path = repo.get('path')
+
+    try:
+      ivy = Bootstrapper.default_ivy()
+    except Bootstrapper.Error as e:
+      raise TaskError('Failed to push {0}! {1}'.format(pushdb_coordinate(jar, entry), e))
+
+    ivysettings = self.generate_ivysettings(ivy, published, publish_local=path)
+    args = [
+      '-settings', ivysettings,
+      '-ivy', ivyxml_path,
+      '-deliverto', '%s/[organisation]/[module]/ivy-[revision].xml' % self.workdir,
+      '-publish', resolver,
+      '-publishpattern', '%s/[organisation]/[module]/'
+                         '[artifact]-[revision](-[classifier]).[ext]' % self.workdir,
+      '-revision', entry.version().version(),
+      '-m2compatible',
+    ]
+
+    if LogOptions.stderr_log_level() == logging.DEBUG:
+      args.append('-verbose')
+
+    if self.local_snapshot:
+      args.append('-overwrite')
+
+    try:
+      ivy.execute(jvm_options=jvm_args, args=args,
+                  workunit_factory=self.context.new_workunit, workunit_name='jar-publish')
+    except Ivy.Error as e:
+      raise TaskError('Failed to push {0}! {1}'.format(pushdb_coordinate(jar, entry), e))
+
   def execute(self):
     self.check_clean_master(commit=(not self.dryrun and self.commit))
 
@@ -531,11 +603,17 @@ class JarPublish(JarTask, ScmPublish):
       dbfile = tgt.provides.repo.push_db(tgt)
       result = pushdbs.get(dbfile)
       if not result:
-        # Create an empty db file if none exists.
-        touch(dbfile)
-
-        db = PushDb.load(dbfile)
-        repo = self.repos[tgt.provides.repo.name]
+        # Create an empty pushdb if no dbfile exists.
+        if (os.path.exists(dbfile)):
+          db = PushDb.load(dbfile)
+        else:
+          safe_mkdir(os.path.dirname(dbfile))
+          db = PushDb()
+        try:
+          repo = self.repos[tgt.provides.repo.name]
+        except KeyError:
+          msg = "Repository %s has no entry in pants.ini's %s section under key 'repos'"
+          raise TaskError(msg % (tgt.provides.repo.name, self._CONFIG_SECTION))
         result = (db, dbfile, repo)
         pushdbs[dbfile] = result
       return result
@@ -544,45 +622,99 @@ class JarPublish(JarTask, ScmPublish):
       return get_db(tgt)[0]
 
     def fingerprint_internal(tgt):
-      pushdb, _, _ = get_db(tgt)
+      pushdb = get_pushdb(tgt)
       entry = pushdb.get_entry(tgt)
       return entry.fingerprint or '0.0.0'
 
-    def stage_artifact(tgt, jar, version, changelog, confs=None, artifact_ext=''):
+    def stage_artifact(tgt, jar, version, changelog, confs=None, artifact_ext='', extra_confs=None):
       def path(name=None, suffix='', extension='jar'):
         return self.artifact_path(jar, version, name=name, suffix=suffix, extension=extension,
                                   artifact_ext=artifact_ext)
 
-      with safe_open(path(suffix='-CHANGELOG', extension='txt'), 'w') as changelog_file:
-        changelog_file.write(changelog)
+      with safe_open(path(suffix='-CHANGELOG', extension='txt'), 'wb') as changelog_file:
+        changelog_file.write(changelog.encode('utf-8'))
       ivyxml = path(name='ivy', extension='xml')
 
-      IvyWriter(get_pushdb).write(tgt, ivyxml, confs=confs)
+      IvyWriter(get_pushdb).write(tgt, ivyxml, confs=confs, extra_confs=extra_confs)
       PomWriter(get_pushdb).write(tgt, path(extension='pom'))
 
       return ivyxml
 
-    def copy_artifact(tgt, jar, version, typename, suffix='', artifact_ext=''):
-      genmap = self.context.products.get(typename)
-      for basedir, jars in genmap.get(tgt).items():
-        for artifact in jars:
-          path = self.artifact_path(jar, version, suffix=suffix, artifact_ext=artifact_ext)
-          safe_mkdir(os.path.dirname(path))
-          shutil.copy(os.path.join(basedir, artifact), path)
-
     def stage_artifacts(tgt, jar, version, changelog):
-      copy_artifact(tgt, jar, version, typename='jars')
+      DEFAULT_IVY_TYPE = 'jar'
+      DEFAULT_CLASSIFIER = ''
+      DEFAULT_EXTENSION = 'jar'
+
+      self._copy_artifact(tgt, jar, version, typename='jars')
       self.create_source_jar(tgt, jar, version)
       doc_jar = self.create_doc_jar(tgt, jar, version)
 
       confs = set(repo['confs'])
+      extra_confs = []
+
+      # Process any extra jars that might have been previously generated for this target, or a
+      # target that it was derived from.
+      publish_extras = self.context.config.getdict(self._CONFIG_SECTION, 'publish_extras') or {}
+      for extra_product in publish_extras:
+        extra_config = publish_extras[extra_product]
+
+        override_name = jar.name
+        if 'override_name' in extra_config:
+          # If the supplied string has a '{target_provides_name}' in it, replace it with the
+          # current jar name. If not, the string will be taken verbatim.
+          override_name = extra_config['override_name'].format(target_provides_name=jar.name)
+
+        classifier = DEFAULT_CLASSIFIER
+        suffix = ''
+        ivy_type = DEFAULT_IVY_TYPE
+        if 'classifier' in extra_config:
+          classifier = extra_config['classifier']
+          suffix = "-{0}".format(classifier)
+          ivy_type = classifier
+
+        extension = DEFAULT_EXTENSION
+        if 'extension' in extra_config:
+          extension = extra_config['extension']
+          if ivy_type == DEFAULT_IVY_TYPE:
+            ivy_type = extension
+
+        # A lot of flexibility is allowed in naming the extra artifact. Because the name must be
+        # unique, some extra logic is required to ensure that the user supplied at least one
+        # non-default value (thus ensuring a uniquely-named artifact in the end).
+        if override_name == jar.name and classifier == DEFAULT_CLASSIFIER and extension == DEFAULT_EXTENSION:
+          raise TaskError("publish_extra for '{0}' most override one of name, classifier or "
+                          "extension with a non-default value.".format(extra_product))
+
+        ivy_tmpl_key = "publish_extra-{0}{1}{2}".format(override_name, classifier, extension)
+
+        # Build a list of targets to check. This list will consist of the current target, plus the
+        # entire derived_from chain.
+        target_list = [tgt]
+        target = tgt
+        while target.derived_from != target:
+          target_list.append(target.derived_from)
+          target = target.derived_from
+        for cur_tgt in target_list:
+          if self.context.products.get(extra_product).has(cur_tgt):
+            self._copy_artifact(cur_tgt, jar, version, typename=extra_product,
+                                suffix=suffix, extension=extension,
+                                override_name=override_name)
+            confs.add(ivy_tmpl_key)
+            # Supply extra data about this jar into the Ivy template, so that Ivy will publish it
+            # to the final destination.
+            extra_confs.append({'name': override_name,
+                                'type': ivy_type,
+                                'conf': ivy_tmpl_key,
+                                'classifier': classifier,
+                                'ext': extension})
+
       confs.add(IvyWriter.SOURCES_CONFIG)
       # don't request docs unless they are available for all transitive targets
       # TODO: doc products should be checked by an independent jar'ing task, and
       # conditionally enabled; see https://github.com/pantsbuild/pants/issues/568
-      if doc_jar and self._java_doc(target) and self._scala_doc(target):
+      if doc_jar and self._java_doc(tgt) and self._scala_doc(tgt):
         confs.add(IvyWriter.JAVADOC_CONFIG)
-      return stage_artifact(tgt, jar, version, changelog, confs)
+      return stage_artifact(tgt, jar, version, changelog, confs, extra_confs=extra_confs)
 
     if self.overrides:
       print('Publishing with revision overrides:\n  %s' % '\n  '.join(
@@ -631,13 +763,13 @@ class JarPublish(JarTask, ScmPublish):
 
       if no_changes and not self.force:
         print('No changes for {0}'.format(pushdb_coordinate(jar, oldentry)))
-        stage_artifacts(target, jar, (newentry.version() if self.force else oldentry.version()).version(), changelog)
+        stage_artifacts(target, jar, oldentry.version().version(), changelog)
       elif skip:
         print('Skipping %s to resume at %s' % (
           jar_coordinate(jar, (newentry.version() if self.force else oldentry.version()).version()),
           coordinate(self.restart_at[0], self.restart_at[1])
         ))
-        stage_artifacts(target, jar, oldver.version(), changelog)
+        stage_artifacts(target, jar, oldentry.version().version(), changelog)
       else:
         if not self.dryrun:
           # Confirm push looks good
@@ -647,13 +779,8 @@ class JarPublish(JarTask, ScmPublish):
             print('\nChanges for %s since %s @ %s:\n\n%s' % (
               coordinate(jar.org, jar.name), oldentry.version(), oldentry.sha, changelog
             ))
-          if os.isatty(sys.stdin.fileno()):
-            push = raw_input('Publish %s with revision %s ? [y|N] ' % (
-              coordinate(jar.org, jar.name), newentry.version()
-            ))
-            print('\n')
-            if push.strip().lower() != 'y':
-              raise TaskError('User aborted push')
+          if not self.confirm_push(coordinate(jar.org, jar.name), newentry.version()):
+            raise TaskError('User aborted push')
 
         pushdb.set_entry(target, newentry)
         ivyxml = stage_artifacts(target, jar, newentry.version().version(), changelog)
@@ -661,53 +788,7 @@ class JarPublish(JarTask, ScmPublish):
         if self.dryrun:
           print('Skipping publish of {0} in test mode.'.format(pushdb_coordinate(jar, newentry)))
         else:
-          resolver = repo['resolver']
-          path = repo.get('path')
-
-          # Get authentication for the publish repo if needed
-          jvm_args = self._jvmargs
-          if repo.get('auth'):
-            user = repo.get('username')
-            password = repo.get('password')
-            if user and password:
-              jvm_args.append('-Dlogin=%s' % user)
-              jvm_args.append('-Dpassword=%s' % password)
-            else:
-              raise TaskError('Unable to publish to %s. %s' %
-                              (repo['resolver'], repo.get('help', '')))
-
-          # Do the publish
-          def publish(ivyxml_path):
-            try:
-              ivy = Bootstrapper.default_ivy()
-            except Bootstrapper.Error as e:
-              raise TaskError('Failed to push {0}! {1}'.format(pushdb_coordinate(jar, newentry), e))
-
-            ivysettings = self.generate_ivysettings(ivy, published, publish_local=path)
-            args = [
-              '-settings', ivysettings,
-              '-ivy', ivyxml_path,
-              '-deliverto', '%s/[organisation]/[module]/ivy-[revision].xml' % self.workdir,
-              '-publish', resolver,
-              '-publishpattern', '%s/[organisation]/[module]/'
-                                 '[artifact]-[revision](-[classifier]).[ext]' % self.workdir,
-              '-revision', newentry.version().version(),
-              '-m2compatible',
-            ]
-
-            if LogOptions.stderr_log_level() == logging.DEBUG:
-              args.append('-verbose')
-
-            if self.local_snapshot:
-              args.append('-overwrite')
-
-            try:
-              ivy.execute(jvm_options=jvm_args, args=args,
-                          workunit_factory=self.context.new_workunit, workunit_name='jar-publish')
-            except Ivy.Error as e:
-              raise TaskError('Failed to push {0}! {1}'.format(pushdb_coordinate(jar, newentry), e))
-
-          publish(ivyxml)
+          self.publish(ivyxml, jar=jar, entry=newentry, repo=repo, published=published)
 
           if self.commit:
             org = jar.org
@@ -723,8 +804,32 @@ class JarPublish(JarTask, ScmPublish):
             )
 
             pushdb.dump(dbfile)
-            self.commit_push(coordinate(org, name, rev))
-            self.scm.refresh()
+            self.commit_pushdb(coordinate(org, name, rev))
+            scm_exception = None
+            for attempt in range(self.get_options().scm_push_attempts):
+              try:
+                self.context.log.debug("Trying scm push")
+                self.scm.push()
+                break # success
+              except Scm.RemoteException as scm_exception:
+                self.context.log.debug("Scm push failed, trying to refresh")
+                # This might fail in the event that there is a real conflict, throwing
+                # a Scm.LocalException (in case of a rebase failure) or a Scm.RemoteException
+                # in the case of a fetch failure.  We'll directly raise a local exception,
+                # since we can't fix it by retrying, but if we do, we want to display the
+                # remote exception that caused the refresh as well just in case the user cares.
+                # Remote exceptions probably indicate network or configuration issues, so
+                # we'll let them propagate
+                try:
+                  self.scm.refresh(leave_clean=True)
+                except Scm.LocalException as local_exception:
+                  exc = traceback.format_exc(scm_exception)
+                  self.context.log.debug("SCM exception while pushing: %s" % exc)
+                  raise local_exception
+
+            else:
+              raise scm_exception
+
             self.scm.tag('%(org)s-%(name)s-%(rev)s' % args,
                          message='Publish of %(coordinate)s initiated by %(user)s %(cause)s' % args)
 
@@ -738,27 +843,27 @@ class JarPublish(JarTask, ScmPublish):
 
   def check_targets(self, targets):
     invalid = defaultdict(lambda: defaultdict(set))
-    derived_by_target = dict()
+    derived_by_target = defaultdict(set)
 
-    def collect(publish_target, walked_target):
-      derived_by_target[walked_target.derived_from] = walked_target
+    def collect_invalid(publish_target, walked_target):
+      for derived_target in walked_target.derived_from_chain:
+        derived_by_target[derived_target].add(walked_target)
       if not walked_target.has_sources() or not walked_target.sources_relative_to_buildroot():
         invalid[publish_target][walked_target].add('No sources.')
       if not walked_target.is_exported:
-        invalid[publish_target][walked_target].add('Does not provide an artifact.')
+        invalid[publish_target][walked_target].add('Does not provide a binary artifact.')
 
     for target in targets:
-      target.walk(functools.partial(collect, target), predicate=lambda t: isinstance(t, Jarable))
+      target.walk(functools.partial(collect_invalid, target),
+                  predicate=lambda t: isinstance(t, Jarable))
 
     # When walking the graph of a publishable target, we may encounter families of sibling targets
     # that form a derivation chain.  As long as one of these siblings is publishable, we can
     # proceed and publish a valid graph.
-    # TODO(John Sirois): This does not actually handle derivation chains longer than 2 with the
-    # exported item in the most derived position - fix this.
     for publish_target, invalid_targets in list(invalid.items()):
       for invalid_target, reasons in list(invalid_targets.items()):
-        derived_target = derived_by_target[invalid_target]
-        if derived_target not in invalid_targets:
+        derived_from_set = derived_by_target[invalid_target]
+        if derived_from_set - set(invalid_targets.keys()):
           invalid_targets.pop(invalid_target)
       if not invalid_targets:
         invalid.pop(publish_target)
@@ -792,7 +897,7 @@ class JarPublish(JarTask, ScmPublish):
             for synthetic in generated:
               yield synthetic
 
-      # Handle the case where a code gen target is in the listed roots and the thus the publishable
+      # Handle the case where a code gen target is in the listed roots and thus the publishable
       # target is a synthetic twin generated by a code gen task upstream.
       for candidate in self.context.target_roots:
         candidates.update(get_synthetic('java', candidate))
@@ -827,12 +932,12 @@ class JarPublish(JarTask, ScmPublish):
     return sha.hexdigest()
 
   def changelog(self, target, sha):
-    return self.scm.changelog(from_commit=sha,
-                              files=target.sources_relative_to_buildroot())
+    return ensure_text(self.scm.changelog(from_commit=sha,
+                                          files=target.sources_relative_to_buildroot()))
 
   def generate_ivysettings(self, ivy, publishedjars, publish_local=None):
     if ivy.ivy_settings is None:
-      raise TaskError('A custom ivysettings.xml with writeable resolvers is required for'
+      raise TaskError('A custom ivysettings.xml with writeable resolvers is required for '
                       'publishing, but none was configured.')
     template_relpath = os.path.join('templates', 'jar_publish', 'ivysettings.mustache')
     template = pkgutil.get_data(__name__, template_relpath)
