@@ -22,6 +22,9 @@ class BuildGraph(object):
   """A directed acyclic graph of Targets and dependencies. Not necessarily connected.
   """
 
+  class DuplicateAddressError(AddressLookupError):
+    """The same address appears multiple times in a dependency list"""
+
   class TransitiveLookupError(AddressLookupError):
     """Used to append the current node to the error message from an AddressLookupError """
 
@@ -158,8 +161,8 @@ class BuildGraph(object):
                      .format(dependent=dependent, dependency=dependency))
 
     if dependency in self.dependencies_of(dependent):
-      logger.warn('{dependent} already depends on {dependency}'
-                  .format(dependent=dependent, dependency=dependency))
+      logger.debug('{dependent} already depends on {dependency}'
+                   .format(dependent=dependent, dependency=dependency))
     else:
       self._target_dependencies_by_address[dependent].add(dependency)
       self._target_dependees_by_address[dependency].add(dependent)
@@ -168,8 +171,6 @@ class BuildGraph(object):
     """Returns all the targets in the graph in no particular order.
 
     :param predicate: A target predicate that will be used to filter the targets returned.
-
-    :return: a list of targets evaluated in inorder traversal order.
     """
     return filter(predicate, self._target_by_address.values())
 
@@ -177,12 +178,14 @@ class BuildGraph(object):
     """:return: targets ordered from most dependent to least."""
     return sort_targets(self._target_by_address.values())
 
-  def walk_transitive_dependency_graph(self, addresses, work, predicate=None):
+  def walk_transitive_dependency_graph(self, addresses, work, predicate=None, postorder=False):
     """Given a work function, walks the transitive dependency closure of `addresses`.
 
     :param list<Address> addresses: The closure of `addresses` will be walked.
     :param function work: The function that will be called on every target in the closure using
-      inorder traversal order.
+      the specified traversal order.
+    :param bool postorder: When ``True``, the traversal order is postorder (children before
+      parents), else it is preorder (parents before children).
     :param function predicate: If this parameter is not given, no Targets will be filtered
       out of the closure.  If it is given, any Target which fails the predicate will not be
       walked, nor will its dependencies.  Thus predicate effectively trims out any subgraph
@@ -194,14 +197,18 @@ class BuildGraph(object):
         walked.add(address)
         target = self._target_by_address[address]
         if not predicate or predicate(target):
-          work(target)
+          if not postorder:
+            work(target)
           for dep_address in self._target_dependencies_by_address[address]:
             _walk_rec(dep_address)
+          if postorder:
+            work(target)
     for address in addresses:
       _walk_rec(address)
 
-  def walk_transitive_dependee_graph(self, addresses, work, predicate=None):
-    """Identical to `walk_transitive_dependency_graph`, but walks dependees inorder traversal order.
+  def walk_transitive_dependee_graph(self, addresses, work, predicate=None, postorder=False):
+    """Identical to `walk_transitive_dependency_graph`, but walks dependees preorder (or postorder
+    if the postorder parameter is True).
 
     This is identical to reversing the direction of every arrow in the DAG, then calling
     `walk_transitive_dependency_graph`.
@@ -212,13 +219,16 @@ class BuildGraph(object):
         walked.add(address)
         target = self._target_by_address[address]
         if not predicate or predicate(target):
-          work(target)
+          if not postorder:
+            work(target)
           for dep_address in self._target_dependees_by_address[address]:
             _walk_rec(dep_address)
+          if postorder:
+            work(target)
     for address in addresses:
       _walk_rec(address)
 
-  def transitive_dependees_of_addresses(self, addresses, predicate=None):
+  def transitive_dependees_of_addresses(self, addresses, predicate=None, postorder=False):
     """Returns all transitive dependees of `address`.
 
     Note that this uses `walk_transitive_dependee_graph` and the predicate is passed through,
@@ -229,10 +239,10 @@ class BuildGraph(object):
     :param function predicate: The predicate passed through to `walk_transitive_dependee_graph`.
     """
     ret = OrderedSet()
-    self.walk_transitive_dependee_graph(addresses, ret.add, predicate=predicate)
+    self.walk_transitive_dependee_graph(addresses, ret.add, predicate=predicate, postorder=False)
     return ret
 
-  def transitive_subgraph_of_addresses(self, addresses, predicate=None):
+  def transitive_subgraph_of_addresses(self, addresses, predicate=None, postorder=False):
     """Returns all transitive dependencies of `address`.
 
     Note that this uses `walk_transitive_dependencies_graph` and the predicate is passed through,
@@ -244,7 +254,9 @@ class BuildGraph(object):
       `walk_transitive_dependencies_graph`.
     """
     ret = OrderedSet()
-    self.walk_transitive_dependency_graph(addresses, ret.add, predicate=predicate)
+    self.walk_transitive_dependency_graph(addresses, ret.add,
+                                          predicate=predicate,
+                                          postorder=postorder)
     return ret
 
   def inject_synthetic_target(self,
@@ -279,67 +291,73 @@ class BuildGraph(object):
                          **kwargs)
     self.inject_target(target, dependencies=dependencies, derived_from=derived_from)
 
-  def inject_address(self, address):
-    """Delegates to an internal AddressMapper to resolve, construct, and inject a Target.
-
-    :param Address address: The address to inject.  Must be resolvable by `self._address_mapper`.
-    """
-    if not self.contains_address(address):
-      target_addressable = self._address_mapper.resolve(address)
-      target = self.target_addressable_to_target(address, target_addressable)
-      self.inject_target(target)
-
   def inject_address_closure(self, address):
-    """Recursively calls `inject_address` through the transitive closure of dependencies."""
+    """Resolves, constructs and injects a Target and its transitive closure of dependencies.
+
+    This method is idempotent and will short circuit for already injected addresses. For all other
+    addresses though, it delegates to an internal AddressMapper to resolve item the address points
+    to.
+
+    :param Address address: The address to inject.  Must be resolvable by `self._address_mapper` or
+                            else be the address of an already injected entity.
+    """
+
+    if self.contains_address(address):
+      # The address was either mapped in or synthetically injected already.
+      return
 
     if address in self._addresses_already_closed:
+      # We've visited this address already in the course of the active recursive injection.
       return
 
     mapper = self._address_mapper
 
-    target_addressable = mapper.resolve(address)
+    target_address, target_addressable = mapper.resolve(address)
 
-    self._addresses_already_closed.add(address)
+    self._addresses_already_closed.add(target_address)
     try:
       dep_addresses = list(mapper.specs_to_addresses(target_addressable.dependency_specs,
-                                                     relative_to=address.spec_path))
+                                                     relative_to=target_address.spec_path))
+      deps_seen = set()
       for dep_address in dep_addresses:
+        if dep_address in deps_seen:
+          raise self.DuplicateAddressError(
+            'Addresses in dependencies must be unique. \'{spec}\' is referenced more than once.'
+            .format(spec=dep_address.spec))
+        deps_seen.add(dep_address)
         self.inject_address_closure(dep_address)
 
-      if not self.contains_address(address):
-        target = self.target_addressable_to_target(address, target_addressable)
+      if not self.contains_address(target_address):
+        target = self._target_addressable_to_target(target_address, target_addressable)
         self.inject_target(target, dependencies=dep_addresses)
       else:
-        target = self.get_target(address)
+        for dep_address in dep_addresses:
+          if dep_address not in self.dependencies_of(target_address):
+            self.inject_dependency(target_address, dep_address)
+        target = self.get_target(target_address)
+
+      def inject_spec_closure(spec):
+        address = mapper.spec_to_address(spec, relative_to=target_address.spec_path)
+        self.inject_address_closure(address)
 
       for traversable_spec in target.traversable_dependency_specs:
-        self.inject_spec_closure(spec=traversable_spec, relative_to=address.spec_path)
+        inject_spec_closure(traversable_spec)
         traversable_spec_target = self.get_target_from_spec(traversable_spec,
-                                                            relative_to=address.spec_path)
+                                                            relative_to=target_address.spec_path)
         if traversable_spec_target not in target.dependencies:
           self.inject_dependency(dependent=target.address,
                                  dependency=traversable_spec_target.address)
           target.mark_transitive_invalidation_hash_dirty()
 
       for traversable_spec in target.traversable_specs:
-        self.inject_spec_closure(spec=traversable_spec, relative_to=address.spec_path)
+        inject_spec_closure(traversable_spec)
         target.mark_transitive_invalidation_hash_dirty()
 
     except AddressLookupError as e:
       raise self.TransitiveLookupError("{message}\n  referenced from {spec}"
-                                       .format(message=e, spec=address.spec))
+                                       .format(message=e, spec=target_address.spec))
 
-
-  def inject_spec_closure(self, spec, relative_to=''):
-    """Constructs a SyntheticAddress from `spec` and calls `inject_address_closure`.
-
-    :param string spec: A Target spec
-    :param string relative_to: The spec_path of the BUILD file this spec was read from.
-    """
-    address = self._address_mapper.spec_to_address(spec, relative_to=relative_to)
-    self.inject_address_closure(address)
-
-  def target_addressable_to_target(self, address, addressable):
+  def _target_addressable_to_target(self, address, addressable):
     """Realizes a TargetAddressable into a Target at `address`.
 
     :param TargetAddressable addressable:
